module MnoEnterprise
  class Nimbus::V4::BaseController < ApplicationController
    before_filter :authenticate_remote_client!
    before_filter :check_parent_resource
    around_filter :prepare_and_handle_error
    skip_before_action :verify_authenticity_token

    class ResponseCode
      attr_accessor :error_code, :http_code

      def initialize(error_code, http_code)
        @error_code = error_code
        @http_code = http_code
      end
    end

    INTERNAL_ERROR = ResponseCode.new('unknownError', 500)
    RESOURCE_NOT_FOUND = ResponseCode.new('notFound', 404)
    PARENT_RESOURCE_NOT_FOUND = ResponseCode.new('parentNotFound', 404)
    AUTHENTICATION_ERROR = ResponseCode.new('notAuthenticated', 401)
    VALIDATION_ERROR = ResponseCode.new('validationError', 400)
    NO_ERROR = ResponseCode.new('noError', 200)

    respond_to :json

    # GET /api/nimbus/v4/<resources_name>
    # Get all resources
    def index
      render_response(get_resources)
    end

    # GET /api/nimbus/v4/:id
    # Show specific resource
    def show
      @resource = get_resource(params[:id])

      if @resource
        render_response(@resource)
        return
      else
        render_error RESOURCE_NOT_FOUND, 'Resource not found: ' + params[:id]
        return
      end
    end

    # POST /api/nimbus/v4/<resources_name>
    # Create new resource and return representation
    def create
      unless params[:data]
        render_error VALIDATION_ERROR, 'Body does not contain the field: data'
        return false
      end

      attributes = params[:data].select { |k, v| params_create_whitelist.include?(k.to_sym) }

      @resource = build_resource(attributes)

      # Save resource or set error code and status
      unless @resource.save
        render_errors VALIDATION_ERROR, @resource.errors
        return
      end

      after_create_resource(@resource)

      render_response(@resource, 201)
    end

    # PUT /api/nimbus/v4/<resources_name>/:id
    # Change, suspend or reactivate a resource
    # Operation passed in body: {"operation":"suspend"}
    def update
      attributes = (params[:data] || {}).select { |k, v| params_update_whitelist.include?(k.to_sym) }

      @resource = get_resource(params[:id])

      # Check resource exists
      unless @resource
        render_error RESOURCE_NOT_FOUND, 'Resource not found: ' + params[:id]
        return false
      end

      if attributes.any?
        update_resource(@resource, attributes)
        unless @resource.save
          render_errors VALIDATION_ERROR, @resource.errors
          return false
        end
      end

      after_update_resource(@resource, attributes)

      render_response(@resource)
    end


    # DELETE /api/nimbus/v4/<resources_name>/:id
    # Delete a resource
    def destroy
      @resource = get_resource(params[:id])
      # Check resource exists
      unless @resource
        render_error RESOURCE_NOT_FOUND, 'Resource not found: ' + params[:id]
        return
      end

      destroy_resource(@resource)
      render_response(@resource)
    end

    #=======================================
    # Private
    #=======================================
    private

    # Retrieve the parent resource in a before filter
    def parent
      @parent ||= MnoEnterprise::Organization.find_by({uid: params[:organization_id]})
    end

    def check_parent_resource
      if params[:organization_id] && !parent
        render_error PARENT_RESOURCE_NOT_FOUND, 'Parent resource not found: ' + params[:organization_id]
        return false
      end
      true
    end

    # Around filter
    # Return properly formatted json message on error
    def prepare_and_handle_error
      begin
        yield
      rescue Exception => e
        logger.error e
        render_error INTERNAL_ERROR, 'Internal error'
      end
    end

    # Read the token from the http header and
    # check that it matches the Westcon auth settings
    def authenticate_remote_client!
      valid = false

      # Check authentication
      authenticate_with_http_basic do |login, password|
        valid = Settings.westcon.api_login == login && Settings.westcon.api_pwd == password
      end

      # Block access in production (for the moment)
      if (valid || Rails.env.development?)
        return true
      else
        render_error AUTHENTICATION_ERROR, 'Authentication error'
        return true
      end
    end

    # Define the list of parameters for creating a resource
    def params_create_whitelist
      []
    end

    # Define the list of parameters for updating a resource
    # Returns params_create_whitelist by default
    def params_update_whitelist
      params_create_whitelist
    end


    # Render an error message with the provided
    # code and header
    def render_error(response_code, message)
      errors = [{
                    id: UUID.new.generate,
                    code: response_code.error_code.underscore.dasherize,
                    title: message,
                    detail: message
                }]
      render json: {error: response_code.error_code, result: {errors: errors}}, status: response_code.http_code
    end


    def render_errors(response_code, errors)
      render({json: {error: response_code.error_code, result: {errors: process_errors(errors)}}, status: response_code.http_code})
    end


    def process_errors(errors)
      raise 'call process_error for a single error' if errors.kind_of?(String)
      # error may be either an array of string, either an iterable of error and error_val
      # In the first case, error_val will  be nil
      errors.map do |error, error_val|
        field = error_val ? error : nil
        label = [error, error_val].compact.join(' ')
        code = label.parameterize.gsub(/\-\d+/, '')
        {
            id: UUID.new.generate,
            code: code,
            title: label,
            detail: label,
            attribute: field,
            value: error_val
        }
      end
    end

    def resource_klass
      raise NotImplemented
    end

    def hash_for(resource)
      raise NotImplemented
    end

    # Build a resource for creation
    # Give unlimited free trial
    def build_resource(attributes)
      attrs = attributes.dup
      res = resource_klass.new
      res.assign_attributes(attrs)
      res
    end

    # Update a resource
    def update_resource(resource, attributes)
      resource.assign_attributes(attributes.select { |k, v| params_update_whitelist.include?(k.to_sym) })
    end

    # TODO: Call some sort of "terminate" method on Organization
    # model which also terminates all the related app_instances
    # Destroy a resource
    def destroy_resource(resource)
      resource.destroy
    end

    # Hook called after the resource has been
    # successfully created (optional)
    def after_create_resource(resource, opts = {})
    end

    # Hook called after the resource has been
    # successfully updated (optional)
    def after_update_resource(resource, opts = {})
    end

    # Render a resource or resources view
    def render_response(resource_or_list, http_code = 200)
      resp = resource_or_list.kind_of?(Array) ? resource_or_list.map { |e| hash_for(e) } : hash_for(resource_or_list)
      render json: {result: resp, error: NO_ERROR.error_code}, status: http_code
    end

    def hash_for_organization(resource)
      {
          id: resource.uid,
          name: resource.name,
          status: resource.suspended? ? 'suspended' : 'active',
      }
    end

    def hash_for_user(resource)
      {
          id: resource.uid,
          name: resource.name,
          surname: resource.surname,
          email: resource.email,
          role: resource.role
      }
    end

    def hash_for_app_instance(resource)
      {
          id: resource.uid,
          product: resource.app.nid,
          status: resource.status,
          stack: resource.stack,
          created_at: resource.created_at
      }
    end

  end
end
