module MnoEnterprise
  class Nimbus::V4::ResourceInstancesController < Nimbus::V4::BaseController

    # POST /api/nimbus/v4/resource_instances
    # Create new resource and return representation
    def create

      unless attributes = params[:data]
        render_error VALIDATION_ERROR, 'Body does not contain the field: data'
        return false
      end

      unless attributes['organization']
        render_error VALIDATION_ERROR, 'Data does not contains the fields: organization'
        return false
      end
      users = attributes['users']
      unless users && users.any?
        render_error VALIDATION_ERROR, 'Resource should contain at least one user'
        return false
      end

      users_validation_errors = []
      users.each do |hash|
        email = hash[:email]
        user = MnoEnterprise::User.find_by(email: email)
        users_validation_errors << "User [#{user.uid}] already exist with email: #{email}" if user
      end
      if users_validation_errors.any?
        render_errors VALIDATION_ERROR, users_validation_errors
        return false
      end
      # Create organization
      organization = provision_organization(attributes['organization'])
      if organization.errors.any?
        render_errors VALIDATION_ERROR, organization.errors
        return false
      end

      # Create users - handle single object and array
      attributes['users'].each_with_index do |hash, index|
        hash = hash.merge('organization' => organization)
        # Create the first user as a Super Admin
        hash['role'] = 'Super Admin' if index == 0
        provision_user(hash)
      end

      # Create app_instances - handle single object and array
      [attributes['app_instances']].flatten.compact.each do |hash|
        organization.app_instances.create hash
      end

      after_create_resource(organization)

      render_response(organization.reload, 201)
    end

    private

      def resource_klass
        Organization
      end

      # Retrieve all organizations
      def get_resources
        MnoEnterprise::Organization.all
      end

      # Retrieve an organization via uid
      def get_resource(id)
        MnoEnterprise::Organization.find_by({uid: id})
      end

      # Creation attributes for users
      def user_params_create_whitelist
        [:name, :surname, :email, :role, :organization]
      end

      # Creation attributes for organizations
      def organization_params_create_whitelist
        [:name]
      end

      def provision_organization(attributes)
        attrs = attributes.select { |k, v| organization_params_create_whitelist.include?(k.to_sym) }

        res = MnoEnterprise::Organization.new(attrs)
        res.free_trial_end_at = 15.years.from_now

        res.save
        res
      end

      def provision_user(attributes)
        attrs = attributes.select { |k, v| user_params_create_whitelist.include?(k.to_sym) }

        res = MnoEnterprise::User.new(attrs.reject { |k, v| [:role, :organization].include?(k.to_sym) })
        res.password = Devise.friendly_token.first(15)
        res.save

        # Update free trial after creation
        res.free_trial_end_at = 15.years.from_now
        res.save

        # Add user to organization
        if attrs['organization']
          args = [res, attrs['role']].compact
          attrs['organization'].add_user(*args)
        end
        res
      end

      def after_create_resource(resource)
        resource.free_trial_end_at = 15.years.from_now
        resource.save
      end

      # Hash representation of the resource
      def hash_for(resource)
        resp = {id: resource.uid}
        # Add organization
        resp[:organization] = hash_for_organization resource
        # Add app_instances
        resp[:app_instances] = resource.app_instances.map do |r|
          hash_for_app_instance r
        end
        # Add users
        resp[:users] = resource.users.map do |r|
          hash_for_user r
        end
        resp
      end
  end
end
